@@ -25,6 +25,9 @@ from skimage.morphology import medial_axis as medial_axis_skimage
 #  - def getInt8Image(img,nsig_std=None)
 #  - def getInt8ImagePyramid(img_list,pad=0,nsig_std=None,bg_val=0)
 #  - def getStdImage(img,sz_kernel=0,flagMedian=0)
+#  - def getShiftedImage(img,dx,dy,c=0)
+#  - def getLoupeImage(img,x1,y1,x2,y2,fac_scale)
+#  - def getMaskedOverlay(img,masks,colors,transp,maskorder=None)
 #
 # Part II) Threshold and Maximum Operations
 #  - def localmaxima(im,rx=3,ry=None,flagMask=1,val0=0)
@@ -195,6 +198,119 @@ def getStdImage(img,sz_kernel=0,flagMedian=0):
                 mean_img,sig_img=0.0,0.0
     return img_std, mean_img, sig_img
 
+def getShiftedImage(img,dx,dy,fillValue=0,fillImage=None):
+    """
+    computed image shifted by dx pixels in x-direction and dy pixels in y-direction; fill empty pixels by c
+    :param img: original image to be shifted
+    :param dx: shift in x-direction
+    :param dy: shift in y-direction
+    :param fillValue: default value for empty array locations
+    :param fillImage: if not None then fill the same locations in fillImage with fillValue as in the shifted image 
+                      such that at filled locations, difference between fillImage and img_shifted will be zero!
+    :returns img_shifted: shifted image   
+    """
+    # preallocate empty array and assign slice by chrisaycock
+    img_shifted = np.empty_like(img)      # preallocate empty array
+    if dx>=0: img_shifted[:,:dx] = fillValue
+    else    : img_shifted[:,dx:] = fillValue
+    if dy>=0: img_shifted[:dy,:] = fillValue
+    else    : img_shifted[dy:,:] = fillValue
+    if not fillImage is None:
+        assert fillImage.shape==img.shape,"fillImage.shape="+str(fillImage.shape)+" must be same as img.shape="+str(img.shape)+"!"
+        if dx>=0: fillImage[:,:dx] = fillValue
+        else    : fillImage[:,dx:] = fillValue
+        if dy>=0: fillImage[:dy,:] = fillValue
+        else    : fillImage[dy:,:] = fillValue
+    if dx>0:
+        if   dy>0: img_shifted[dy:,dx:] = img[:-dy,:-dx]
+        elif dy<0: img_shifted[:dy,dx:] = img[-dy:,:-dx]
+        else     : img_shifted[:  ,dx:] = img[:   ,:-dx]
+    elif dx<0:
+        if   dy>0: img_shifted[dy:,:dx] = img[:-dy,-dx:]
+        elif dy<0: img_shifted[:dy,:dx] = img[-dy:,-dx:]
+        else     : img_shifted[:  ,:dx] = img[:   ,-dx:]
+    else:
+        if   dy>0: img_shifted[dy:,:  ] = img[:-dy,:   ]
+        elif dy<0: img_shifted[:dy,:  ] = img[-dy:,:   ]
+        else     : img_shifted[:  ,:  ] = img[:   ,:   ]
+    return img_shifted
+
+def getLoupeImage(img,pos,r,fac_scale,pad='default',bbox_clipped_region=None): 
+    """
+    compute enlarged image crop from source image 
+    :param img      : original image where the loupe operation should be applied to
+    :param pos      : list [x,y] of central position where to crop from original image  
+    :param r        : radius (or list of radii [rx,ry]) defining the crop region in the original image
+    :param fac_scale: enlargement factor (or list of factors [fac_x,fac_y]) for the loupe image (to be applied to the cropped image)
+    :param pad      : if 'default' then fill border regions with 0; otherwise fill up with pad 
+    :param bbox_clipped_region: if not None, then should be list [x,y,w,h] of length 4; values will be changed (in-place) to bbox of clipped region in img 
+    :returns img_loupe: loupe image (cropped and enlarged)
+    """
+    if not isinstance(r        , (list, tuple)): r=[r,r]
+    if not isinstance(fac_scale, (list, tuple)): fac_scale=[fac_scale,fac_scale]
+    assert len(pos)      ==2,"pos=(x,y) must be list of length 2, but actually pos="+str(pos)
+    assert len(r)        ==2,"r  =(rx,ry) must be list of length 2, but actually r="+str(r)
+    assert len(fac_scale)==2,"fac_scale=(scale_x,scale_y) must be list of length 2, but actually fac_scale="+str(fac_scale)
+    x,y,rx,ry  =pos[0],pos[1],r[0],r[1]     # central position and raddi in x/y direction of crop region
+    x1,y1,x2,y2=x-rx,y-ry,x+rx+1,y+ry+1     # bbox coordinates of crop region
+    sc_x,sc_y  =fac_scale[0],fac_scale[1]   # scaling factors in x/y direction
+    x1_img=min(max(x1,0),img.shape[1])
+    y1_img=min(max(y1,0),img.shape[0])
+    x2_img=min(max(x2,x1+1),img.shape[1])
+    y2_img=min(max(y2,y1+1),img.shape[0])
+    img_loupe_shape = tuple([y2-y1,x2-x1]+list(img.shape[2:]))
+    img_loupe       = np.zeros(img_loupe_shape,dtype=img.dtype)
+    if pad!='default': img_loupe[:,:]=pad
+    img_loupe[(y1_img-y1):(img_loupe_shape[0]+y2_img-y2),(x1_img-x1):(img_loupe_shape[1]+x2_img-x2)]= img[y1_img:y2_img,x1_img:x2_img]        # crop image
+    if not bbox_clipped_region is None:
+        assert isinstance(bbox_clipped_region,list) and len(bbox_clipped_region)==4,"optional parameter bbox_clipped_region="+str(bbox_clipped_region)+" should be either None or bounding box list [x,y,w,h]"
+        bbox_clipped_region[0]=x1_img
+        bbox_clipped_region[1]=y1_img
+        bbox_clipped_region[2]=x2_img-x1_img
+        bbox_clipped_region[3]=y2_img-y1_img
+    sh_img_loupe    = np.array(img_loupe.shape) # shape of cropped image
+    sh_img_loupe[0]*= max(1,sc_y)               # scale shape of loupe image (y-direction)
+    sh_img_loupe[1]*= max(1,sc_x)               # ditto for x direction
+    img_loupe=cv2.resize(img_loupe,(sh_img_loupe[0],sh_img_loupe[1]),interpolation=cv2.INTER_NEAREST)   # scale loupe image
+    return img_loupe
+
+def getMaskedOverlay(img,masks,colors,transp,maskorder=None): 
+    """
+    compute masked overlay image for (possibly multiple) binary masks, each having its own color  
+    :param img        : original (color) image of size [h,w,3] 
+    :param masks      : binary (0/1) mask of size [h,w] or multiple masks of size [nMasks,h,w] 
+    :param colors     : array of RGB values [[R,G,B],...] for each mask (or a single RGB value for all masks)  
+    :param transp     : transparency value between 0 and 1 (0 means plot only labels); (1 means plot only image) 
+    :param maskorder  : list of mask indexes defining in which order to apply colors (only last color will "survive")
+    :returns img_ovrly: overlay image 
+    """
+    assert len(img.shape)==3 and img.shape[2]==3, "Expected RGB image with shape [h,w,3], but obtained img.shape="+str(img.shape)
+    assert len(masks.shape)>=2 and len(masks.shape)<=3, "masks must be either 2D or 3D array, but masks.shape="+str(masks.shape)
+    if len(masks.shape)==2: masks=np.expand_dims(masks,0)    # if only one mask then expand dimension (so that we have an 3D-array of 2D masks)
+    assert masks.shape[1]==img.shape[0] and masks.shape[2]==img.shape[1],"Image must have same size [h,w] as masks! But masks.shape="+str(masks.shape)+" whereas img.shape="+str(img.shape)
+    h,w = img.shape[0], img.shape[1]      # image size
+    nMasks=masks.shape[0]                 # number of masks
+    if maskorder is None: maskorder=range(nMasks)           # default order is 0,1,2,...
+    colors=np.array(colors)
+    if len(colors.shape)==1: np.expand_dims(colors,0)       # if only one color then expand dimension (so that we have an 2D-array of 1D colors)
+    nColors=colors.shape[0]
+    assert nColors==1 or nColors==nMasks,"Number of colors must be either 1 or equal to the number of masks, but nColors="+str(nColors)+" and nMasks="+str(nMasks)
+    img_ovrly=np.array(img,dtype=img.dtype)
+    if transp<1.0:
+        mask1=np.zeros((h,w),dtype='uint8')                   # master mask (OR-ing of all masks)
+        mask1[np.sum(masks[maskorder],axis=0)>0]=1          # set pixels where at least one mask is >0
+        img_mask_col=np.zeros((h,w,3),dtype=img.dtype)      # allocate array for image mask colors (containing color values for each mask pixel)
+        for i in maskorder:
+            assert i>=0 and i<nMasks,"maskorder="+str(maskorder)+" contains value i="+str(i)+" that is either <0 or exceeds number of masks nMasks="+str(nMasks)
+            img_mask_col[masks[i,:,:]>0]=colors[i%nColors]
+        idx_mask_on=mask1>0
+        if transp<=0.0:
+            img_ovrly[idx_mask_on]=img_mask_col[idx_mask_on] # No transparency? Then just use full colors for mask regions
+        else:
+            img_ovrly[idx_mask_on]=transp*img_ovrly[idx_mask_on] + (1.0-transp)*img_mask_col[idx_mask_on]
+    return img_ovrly
+
+    
 # ***************************************************************************************
 # Part II) Threshold and Maximum Operations
 # ***************************************************************************************
